/* hcalproxy -- personal HTTP proxy to convert HTML+hcalendar to icalendar
 *
 * Works as an HTTP server. When presented with a request for
 * http://thismachine:thisport/http://othermachine/path/to/document,
 * it will issue a request to othermachine, parse the result (after
 * testing checking it it is (X)HTML, and return an icalendar
 * (text/calendar) formed from all the hcalendar entries in that
 * docoument.
 *
 * Options:
 *
 * -d
 *
 *     Run as daemon (i.e., in background). By default, the program
 *     runs in the foreground.
 *
 * -p port-or-service
 *
 *     The port number (or the name of a service from /etc/services)
 *     to listen on. Default is 8000.
 *
 * To do:
 *
 * Handle HEAD.
 *
 * Handle hatom and hcard with same code? Maybe use different URL
 * prefixes to distinguish different functions.
 *
 * Add some form of (optional) logging. Use syslog?
 *
 * Created: 4 Aug 2008
 * Author: Bert Bos <bert@w3.org
 *
 * Copyright © 2008 World Wide Web Consortium (Massachusetts Institute
 * of Technology, European Research Consortium for Informatics and
 * Mathematics, Keio University). All Rights Reserved. This work is
 * distributed under the W3C® Software License[1] in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * [1] http://www.w3.org/Consortium/Legal/copyright-software
 */

#define _GNU_SOURCE		/* We want strcasestr() */
#include "config.h"
#if HAVE_SYS_TYPES_H
#  include <sys/types.h>
#endif
#if HAVE_SYS_SOCKET_H
#  include <sys/socket.h>
#endif
#include <sys/wait.h>
#include <sys/errno.h>
#if HAVE_NETINET_IN_H
#  include <netinet/in.h>
#endif
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#if HAVE_UNISTD_H
#  include <unistd.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <assert.h>
#if HAVE_STRINGS_H
#  include <strings.h>
#endif
#if HAVE_STRING_H
#  include <string.h>
#endif
#include <stdarg.h>
#include <ctype.h>
#include <time.h>
#include "heap.e"
#include "types.e"
#include "errexit.e"
#include "scan.e"
#include "html.e"
#include "dict.e"
#include "connectsock.e"
#include "openurl.e"
#include "tree.e"
#include "class.e"
#include "unent.e"
#include "headers.e"
#include "proxy.e"


#define QLEN 5			/* Maximum connection queue length */


/* reaper -- clean up zombie children */
static RETSIGTYPE reaper(int unused)
{
  while (waitpid(-1, NULL, WNOHANG) >= 0) ;
}


int main(int argc, char *argv[])
{
  string port = "8000";		/* Service name or port number to listen on */
  struct sockaddr_in fsin;	/* The address of a client */
  Boolean daemon = False;	/* Whether to run in the background */
  socklen_t alen;		/* Length of client's address */
  int msock;			/* Master server socket */
  int ssock;			/* Slave server socket */
  char c;

  while ((c = getopt(argc, argv, "dp:")) != -1)
    switch (c) {
    case 'd': daemon = True; break;
    case 'p': port = optarg; break;
    default: errexit("Usage: %s [-d] [-p port-or-service-name]\n", argv[0]);
    }

  msock = passiveTCP(port, QLEN);
  if (msock < 0)
    errexit("%s: Could not open port %s for listening: %s\n",
	    argv[0], port, strerror(errno));

  if (daemon) {
    switch (fork()) {
    case -1:			/* Error */
      errexit("%s: fork(): %s\n", argv[0], strerror(errno));
    case 0:			/* Child */
      printf("Going into background. Process ID = %d\n", getpid());
      close(0);
      close(1);
      /* close(2); */
      break;
    default:			/* Parent */
      exit(0);
    }
  }

  (void) signal(SIGCHLD, reaper);

  while (1) {
    alen = sizeof(fsin);
    ssock = accept(msock, (struct sockaddr *)&fsin, &alen);
    if (ssock < 0) {
      if (errno == EINTR) continue;
      errexit("%s: accept(): %s\n", argv[0], strerror(errno));
    }
    switch (fork()) {
    case 0:			/* Child */
      (void) close(msock);
      exit(proxy(ssock));
    default:			/* Parent */
      (void) close(ssock);
      break;
    case -1:
      errexit("%s: fork(): %s\n", argv[0], strerror(errno));
    }
  }
}
