#define _GNU_SOURCE		/* We want strcasestr() */
#include "config.h"
#if HAVE_SYS_TYPES_H
#  include <sys/types.h>
#endif
#if HAVE_SYS_SOCKET_H
#  include <sys/socket.h>
#endif
#include <sys/wait.h>
#include <sys/errno.h>
#if HAVE_NETINET_IN_H
#  include <netinet/in.h>
#endif
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#if HAVE_UNISTD_H
#  include <unistd.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <assert.h>
#if HAVE_STRINGS_H
#  include <strings.h>
#endif
#if HAVE_STRING_H
#  include <string.h>
#endif
#include <stdarg.h>
#include <ctype.h>
#include <time.h>
#include "export.h"
#include "heap.e"
#include "types.e"
#include "errexit.e"
#include "scan.e"
#include "html.e"
#include "dict.e"
#include "connectsock.e"
#include "openurl.e"
#include "tree.e"
#include "class.e"
#include "unent.e"
#include "headers.e"


#define BUFLEN 4096		/* Maximum length of an HTTP header */

#if !HAVE_STRCASECMP
#error To do: implement strcasecmp()
#endif

/* icalendar data types */
typedef enum {P_DATE, P_TEXT, P_LIST, P_URL} PropertyType;

static Tree tree;
Boolean has_errors = False;
Boolean iso_8859_1;


/* debug -- print debugging info */
static void debug(char *format,...)
{
#ifdef DEBUG
  va_list ap;

  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
#endif
}


/* handle_error -- called when a parse error occurred */
void handle_error(void *clientdata, const string s, int lineno)
{
  debug("+ Error on line %d: %s\n", lineno, s);
  has_errors = True;
}


/* start -- called before the first event is reported */
void* start(void)
{
  debug("+ Start parsing\n");
  tree = create();
  return NULL;
}
  

/* end -- called after the last event is reported */
void end(void *clientdata)
{
  debug("+ End parsing\n");
  /* skip */
}


/* handle_comment -- called after a comment is parsed */
void handle_comment(void *clientdata, string commenttext)
{
  /* We don't need comments, otherwise the statement would be: */
  /* tree = append_comment(tree, commenttext); */
}


/* handle_text -- called after a text chunk is parsed */
void handle_text(void *clientdata, string text)
{
  tree = append_text(tree, text);
}


/* handle_decl -- called after a declaration is parsed */
void handle_decl(void *clientdata, string gi, string fpi, string url)
{
  /* We don't need doctypes, otherwise the statement would be: */
  /* tree = append_declaration(tree, gi, fpi, url); */
}


/* handle_pi -- called after a PI is parsed */
void handle_pi(void *clientdata, string pi_text)
{
  /* We don't need processing instructions, otherwise the statement would be: */
  /* tree = html_push(tree, name, attribs); */
}


/* handle_starttag -- called after a start tag is parsed */
void handle_starttag(void *clientdata, string name, pairlist attribs)
{
  tree = html_push(tree, name, attribs);
}


/* handle_endtag -- called after an endtag is parsed (name may be "") */
void handle_endtag(void *clientdata, string name)
{
  tree = html_pop(tree, name);
  dispose(name);
}


/* handle_emptytag -- called after an empty tag is parsed */
void handle_emptytag(void *clientdata, string name, pairlist attribs)
{
  tree = html_push(tree, name, attribs);
}


/* append_utf8 -- append the UTF-8 sequence for code n */
static void append_utf8(const int n, string t, int *j)
{
  if (n <= 0x7F) {
    t[(*j)++] = n;
  } else if (n <= 0x7FF) {
    t[(*j)++] = (char)(0xC0 | (n >> 6));
    t[(*j)++] = (char)(0x80 | (n & 0x3F));
  } else if (n <= 0xFFFF) {
    t[(*j)++] = (char)(0xE0 | (n >> 12));
    t[(*j)++] = (char)(0x80 | ((n >> 6) & 0x3F));
    t[(*j)++] = (char)(0x80 | (n & 0x3F));
  } else if (n <= 0x1FFFFF) {
    t[(*j)++] = (char)(0xF0 | (n >> 18));
    t[(*j)++] = (char)(0x80 | ((n >> 12) & 0x3F));
    t[(*j)++] = (char)(0x80 | ((n >> 6) & 0x3F));
    t[(*j)++] = (char)(0x80 | (n & 0x3F));
  } else if (n <= 0x3FFFFFF) {
    t[(*j)++] = (char)(0xF0 | (n >> 24));
    t[(*j)++] = (char)(0x80 | ((n >> 18) & 0x3F));
    t[(*j)++] = (char)(0x80 | ((n >> 12) & 0x3F));
    t[(*j)++] = (char)(0x80 | ((n >> 6) & 0x3F));
    t[(*j)++] = (char)(0x80 | (n & 0x3F));
  } else {
    t[(*j)++] = (char)(0xF0 | (n >> 30));
    t[(*j)++] = (char)(0x80 | ((n >> 24) & 0x3F));
    t[(*j)++] = (char)(0x80 | ((n >> 18) & 0x3F));
    t[(*j)++] = (char)(0x80 | ((n >> 12) & 0x3F));
    t[(*j)++] = (char)(0x80 | ((n >> 6) & 0x3F));
    t[(*j)++] = (char)(0x80 | (n & 0x3F));
  }
}

/* get_contents -- collect all text content of an elt into a single string */
static string get_contents(Tree t)
{
  Node *h;
  string contents = NULL, k;

  assert(t->tp == Element);
  for (h = t->children; h; h = h->sister) {
    if (h->tp == Text) {
      strapp(&contents, h->text, NULL);
    } else if (h->tp == Element && (k = get_contents(h))) {
      strapp(&contents, k, NULL);
      dispose(k);
    }
  }
  return contents;
}


/* toint -- interpret at most n digits from *s in decimal, advance *s */
static int toint(conststring *s, int n)
{
  int h;

  for (h = 0; n && isdigit(**s); (*s)++, n--) h = 10 * h + **s - '0';
  return h;
}

/* convert_date -- convert ISO date to icalendar date */
static string convert_date(const conststring isodate)
{
  string icaldate;
  long offset = 0;
  conststring s;
  struct tm tm;
  time_t t;

  newarray(icaldate, 17);	/* yyyymmddThhmmssZ */
  for (s = isodate; isspace(*s); s++);
  tm.tm_year = toint(&s, 4) - 1900;
  if (*s == '-') s++;
  tm.tm_mon = toint(&s, 2) - 1;
  if (*s == '-') s++;
  tm.tm_mday = toint(&s, 2);
  if (*s == 'T') s++;
  tm.tm_hour = toint(&s, 2);
  if (*s == ':') s++;
  tm.tm_min = toint(&s, 2);
  if (*s == ':') s++;
  tm.tm_sec = toint(&s, 2);
  if (*s == '+' || *s == '-') {
    Boolean neg = *(s++) == '-';
    offset = 60 * 60 * toint(&s, 2);
    if (*s == ':') s++;
    offset += 60 * toint(&s, 2);
    if (neg) offset = -offset;
  }
  t = timegm(&tm) - offset;
  if (strftime(icaldate, 17, "%Y%m%dT%H%M%SZ", gmtime(&t)) != 0); /* ??? */
  return icaldate;
}


/* decode -- decode an HTML text: convert charset and expand entities */
static string decode(conststring s)
{
  const struct _Entity *e;
  int i, j, n;
  string t;

  if (!(t = malloc((2 * strlen(s) + 1) * sizeof(*s)))) return NULL;
  for (i = 0, j = 0; s[i];) {
    if (s[i] != '&') {				/* Literal character */
      if (s[i] > 0 || !iso_8859_1) t[j++] = s[i++]; /* Don't convert */
      else append_utf8(0xFF & s[i++], t, &j);
    } else if (s[i+1] != '#') {			/* Named entity, eg. &eacute */
      for (i++, n = 0; isalnum(s[i+n]); n++) ;
      if (! (e = lookup_entity(s + i, n))) {	/* Unknown entity */
	t[j++] = '&';
      } else {		 			/* Expand to UTF-8 */
	append_utf8(e->code, t, &j);
	i += n;
	if (s[i] == ';') i++;
      }
    } else if (s[i+2] != 'x') {			/* Decimal entity, eg. &#70 */
      for (n = 0, i += 2; isdigit(s[i]); i++) n = 10 * n + s[i] - '0';
      append_utf8(n, t, &j);
      if (s[i] == ';') i++;
    } else {					/* Hex entity, eg. &#x5F */
      for (n = 0, i += 3; isxdigit(s[i]); i++)
	if (isdigit(s[i])) n = 16 * n + s[i] - '0';
	else n = 16 * n + toupper(s[i]) - 'A' + 10;
      append_utf8(n, t, &j);
      if (s[i] == ';') i++;
    }
  }
  t[j] = '\0';
  return t;
  /* SGML says also that a record-end (i.e., an end-of-line) may be
   * used instead of a semicolon to end an entity reference. But the
   * record-end is not suppressed in HTML and such an entity reference
   * is invalid in XML, so we don't implement that rule here. Instead,
   * the end-of-line is treated as any other character (other than
   * semicolon) and left in the document.
   */
}


/* escape -- escape icalendar reserved characters */
static string escape(const conststring value)
{
  string t;
  int i, j;

  newarray(t, 2 * strlen(value) + 1); /* Worst case is twice as long */
  for (i = 0, j = 0; value[i]; i++) {
    switch (value[i]) {
    case ' ': case '\t': case '\r': case '\v': case '\f':
      if (i != 0 && !isspace(value[i-1])) t[j++] = ' ';
      break;
    case '\\': t[j++] = '\\'; t[j++] = '\\'; break;
    case ';': t[j++] = '\\'; t[j++] = ';'; break;
    case ',': t[j++] = '\\'; t[j++] = ','; break;
    case '\n':
      if (i != 0 && !isspace(value[i-1])) t[j++] = ' ';
      break;
    case '\177': break;
    default: if (0xE0 & value[i]) t[j++] = value[i]; /* Remove control chars */
    }
  }
  t[j] = '\0';
  return t;
}


/* find_prop -- check if element sets a value for property and if so store it */
static void find_prop(PropertyType type, const conststring p,
		      const conststring class, const conststring value,
		      Tree tree, Dictionary d)
{
  conststring v;
  string s, t;

  if (!contains(class, p)) return;
  if (value) t = decode(value);
  else if ((s = get_contents(tree))) {t = decode(s); dispose(s);}
  else return;

  if (type == P_DATE) {
    s = t;
    t = convert_date(t);
    dispose(s);
  } else if (type != P_URL && (v = dict_find(d, p))) {
    s = escape(t);
    dispose(t);
    t = strapp(NULL, v, type == P_LIST ? "," : "", s, NULL);
    dispose(s);
  } else {
    s = t;
    t = escape(t);
    dispose(s);
  }
  dict_add(d, p, t);
  dispose(t);
  /* To do: handle error if dict_add returns 0 */
}


/* parse_event -- extract and print event properties from a subtree */
static void parse_event(Tree t, Dictionary props, Boolean do_includes)
{
  string class, rel, url = NULL, title = NULL, data = NULL;
  Tree h;

  assert(t->tp == Element);
  if (get_attrib(t, "class", &class)) {
    if (strcasecmp(t->name, "abbr") == 0) get_attrib(t, "title", &title);
    if (strcasecmp(t->name, "a") == 0) get_attrib(t, "href", &url);
    if (strcasecmp(t->name, "object") == 0) get_attrib(t, "data", &data);
    /* To do: Can values be spread over disjoint elements? Use a Dictionary? */
    find_prop(P_DATE, "DTSTART", class, title, t, props);
    find_prop(P_TEXT, "SUMMARY", class, title, t, props);
    find_prop(P_TEXT, "LOCATION", class, title, t, props);
    find_prop(P_DATE, "DTEND", class, title, t, props);
    find_prop(P_DATE, "RDATE", class, title, t, props);
    find_prop(P_TEXT, "RRULE", class, title, t, props);
    find_prop(P_LIST, "CATEGORIES", class, title, t, props);
    find_prop(P_TEXT, "DESCRIPTION", class, title, t, props);
    find_prop(P_TEXT, "UID", class, title, t, props);
    find_prop(P_TEXT, "ATTENDEE", class, title, t, props);
    find_prop(P_TEXT, "CONTACT", class, title, t, props);
    find_prop(P_TEXT, "ORGANIZER", class, title, t, props);
    find_prop(P_TEXT, "DURATION", class, title, t, props);
    find_prop(P_TEXT, "CLASS", class, title, t, props);
    find_prop(P_TEXT, "COMMENT", class, title, t, props);
    find_prop(P_TEXT, "GEO", class, title, t, props);
    find_prop(P_TEXT, "PRIORITY", class, title, t, props);
    find_prop(P_TEXT, "RESOURCES", class, title, t, props);
    find_prop(P_TEXT, "STATUS", class, title, t, props);
    find_prop(P_TEXT, "TRANSP", class, title, t, props);
    find_prop(P_URL, "URL", class, url, t, props);
    find_prop(P_URL, "ATTACHMENT", class, url, t, props);
    if (do_includes && contains(class, "include") && hasprefix(data, "#")) {
      /* Should this be recursive instead? How about infinite loops? */
      h = get_elt_by_id(t, data + 1);
      if (h) parse_event(h, props, False);
    }
  }
  if (get_attrib(t, "rel", &rel)) {
    if (strcasecmp(t->name, "a") == 0) get_attrib(t, "href", &url);
    find_prop(P_URL, "REL", rel, url, t, props);
    find_prop(P_URL, "ATTACHMENT", rel, url, t, props);
  }
  for (h = t->children; h != NULL; h = h->sister)
    if (h->tp == Element) parse_event(h, props, do_includes);
}


static void print_props(FILE *f, Dictionary props)
{
  conststring k;

  fprintf(f, "BEGIN:VEVENT\r\n");
  for (k = dict_next(props, NULL); k; k = dict_next(props, k))
    fprintf(f, "%s:%s\r\n", k, dict_find(props, k));
  fprintf(f, "END:VEVENT\r\n\r\n");
}


/* walk_r -- walk recursively through the HTML tree looking for vevent */
static void walk_r(Tree t, FILE *f)
{
  Dictionary props;
  Tree h;

  assert(t->tp == Element);
  if (has_class(t->attribs, "vevent")) {
    props = dict_create(100);
    parse_event(t, props, True);
    print_props(f, props);
    dict_delete(props);
  } else {
    for (h = t->children; h != NULL; h = h->sister)
      if (h->tp == Element) walk_r(h, f);
  }
}

/* walk -- walk through the HTML tree looking for vevent */
static void walk(Tree t, FILE *f)
{
  Tree h;

  assert(t->tp == Root);
  for (h = t->children; h != NULL; h = h->sister)
    if (h->tp == Element) walk_r(h, f);
}


/* http_err -- print HTTP error code and message, return code */
static int http_err(FILE *f, int status, const conststring msg,
		    const conststring headers, const conststring body)
{
  debug("+ %s\n", msg);
  assert(headers && (headers[0] == '\0' || hasaffix(headers, "\r\n")));
  fprintf(f, "HTTP/1.1 %d %s\r\n%s\r\n%s", status, msg, headers, body);
  fclose(f);
  return status;
}


/* proxy -- proxy a request, convert resulting hcalendar to icalendar */
EXPORT int proxy(const int fd)
{
  char buf[BUFLEN], url[BUFLEN];
  FILE *remote, *client;
  string s;
  const char *v;
  int i, j;
  Dictionary responseheaders, requestheaders;

  client = fdopen(fd, "r+");

  /* Read the HTTP request */
  if (! fgets(buf, sizeof(buf), client)) return ferror(client);
  debug("* %s", buf);

  if (!hasprefix(buf, "GET "))
    return http_err(client, 501, "Method not implemented", "", "");

  /* To do: implement the HEAD method */

  /* If the URL starts with http://, skip until after our own server name */
  if (!hasprefix(buf + 4, "http://")) i = 5;
  else if ((s = strchr(buf + 11, '/'))) i = s - buf + 1;
  else return http_err(client, 400, "Error in URL", "", "");

  /* Extract the URL */
  for (j = 0; buf[i] && buf[i] != ' '; i++, j++) url[j] = buf[i];

  /* Check that we understand this HTTP version */
  if (!hasprefix(buf + i + 1, "HTTP/1.0") &&
      !hasprefix(buf + i + 1, "HTTP/1.1"))
    return http_err(client, 505, "HTTP version not supported",
		    "Content-Type: text/plain\r\n",
		    "Only HTTP versions HTTP/1.0 and HTTP/1.1 are supported.");

  debug("+ %s\n", url);

  /* Read the request headers */
  requestheaders = dict_create(50);
  if (!requestheaders)
    return http_err(client, 500, "Internal error: out of memory", "", "");
  if (!read_mail_headers(client, requestheaders))
    return http_err(client, 400, "Error in request headers", "", "");

  shutdown(fd, 0);		/* Close the socket for reading */

  /* Remove the Host header */
  dict_destroy(requestheaders, "host");

  /* Add a Via header */
  v = dict_find(requestheaders, "via");
  s = v ? strapp(NULL, v, ", ", NULL) : NULL;
  strapp(&s, PACKAGE_VERSION, " ", PACKAGE_NAME, NULL);
  if (!dict_add(requestheaders, "via", s))
    return http_err(client, 500, "Internal error: out of memory", "", "");

  debug("+ added Via header\n");

  /* Issue request to remote server */
  responseheaders = dict_create(50);
  if (!responseheaders)
    return http_err(client, 500, "Internal error: out of memory", "", "");

  remote = fopenurl2(url, "r", requestheaders, responseheaders);
  if (!remote) {
    if (errno == EACCES) return http_err(client, 400, "Error in URL", "", "");
    else if (errno == EFAULT) return http_err(client, 404, "Not found", "", "");
    else return http_err(client, 502, "Proxy error", "", "");
  }

  /* Find content type, check if it is (X)HTML */
  v = dict_find(responseheaders, "content-type");
  if (!v) return http_err(client, 502, "Server sent no content-type", "", "");
  if (!hasprefix(v, "text/html") &&
      !hasprefix(v, "application/xhtml+xml"))
    return http_err(client, 502, "Content type was not (X)HTML", "", "");

  debug("+ is (X)HTML\n", s);
  iso_8859_1 = !(s = strcasestr(v, "charset")) || strcasestr(s, "iso-8859-1");
  debug("+ charset assumed %s\n", iso_8859_1 ? "iso-8859-1" : "utf-8");
  /* To do: handle more character encodings, use iconv(3) */

  /* Set up the tokenizer and parser */
  set_error_handler(handle_error);
  set_start_handler(start);
  set_end_handler(end);
  set_comment_handler(handle_comment);
  set_text_handler(handle_text);
  set_decl_handler(handle_decl);
  set_pi_handler(handle_pi);
  set_starttag_handler(handle_starttag);
  set_emptytag_handler(handle_emptytag);
  set_endtag_handler(handle_endtag);
  yyin = remote;
  if (yyparse() != 0 || has_errors)
    return http_err(client, 502, "Errors in remote HTML", "", "");

  /* Print response headers */
  fprintf(client, "HTTP/1.1 200 OK\r\n");
  fprintf(client, "Content-Type: text/calendar;charset=utf-8\r\n");
  if ((v = dict_find(responseheaders, "via")))
    fprintf(client, "Via: %s, %s %s\r\n\r\n", v, PACKAGE_VERSION, PACKAGE_NAME);
  else
    fprintf(client, "Via: %s %s\r\n\r\n", PACKAGE_VERSION, PACKAGE_NAME);

  /* Look for events */
  debug("+ Start walking...\n");
  fprintf(client, "BEGIN:VCALENDAR\r\n");
  fprintf(client, "VERSION:2.0\r\n");
  fprintf(client, "PRODID:-//W3C//NONSGML %s//EN\r\n", PACKAGE_STRING);
  fprintf(client, "METHOD:PUBLISH\r\n");
  fprintf(client, "X-ORIGINAL-URL:%s\r\n", url); /* To do: escape \\ and \, */
  fprintf(client, "\r\n");
  walk(get_root(tree), client);
  fprintf(client, "END:VCALENDAR\r\n");
  fclose(client);
  debug("+ Done!\n");
  return 0;

#if 0
  while ((i = fread(buf, 1, sizeof(buf), remote)) > 0) {
    if (fwrite(buf, 1, i, client) == 0) return ferror(client);
  }
  return ferror(remote);
#endif
}
